/*
           panteltje mcam internet connection routines

         Copyright (C) 1999, 2000  Jan Mourer

 email: jan@panteltje.demon.nl
 snail mail:
 Jan Mourer
 Monnikebildtdijk 2
 9078 VE Oude Bildtzijl
 Holland

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include "vvl_defs.h"
#include "internet.h"
#include "mcam.h"

extern int debug_flag;
extern int accept_internet_connections;
extern BYTE __huge *DataBuffer;
extern BYTE __huge *gStreamBuffer;
extern char *server_password;
extern char *server_control_password;
extern int start_frame, frame_count, frame_delay, save_frames_flag;
extern int save_frames_format;
extern int picture_format;
extern int output_to_frame_flag;
extern char *frame_dir;
extern int store_only_on_change_flag;
extern int coding_mode;
extern int coding_quality;
extern char *home_dir;
extern int gain_ceiling;
extern int man_exposure_flag;
extern int color_balance_mode; /* 1 = auto */
extern char* command_line_preset_filename;
extern int clock_position;
extern int use_utc_flag;
extern int show_time_flag;
extern char screen_text[];
extern int text_to_screen_flag;
extern int text_x, text_y;
extern int text_fr, text_fg, text_fb, text_br, text_bg, text_bb;
extern int selected_font;
extern int apcor_values[];
extern int sensor_frames_per_second;

int make_socket (unsigned short int port)
{
int sock;
struct sockaddr_in name;
     
/* Create the socket. */
sock = socket (PF_INET, SOCK_STREAM, 0);
if (sock < 0)
	{
	perror ("socket");
	return -1;
	}
     
/* Give the socket a name. */
name.sin_family = AF_INET;
name.sin_port = htons (port);
name.sin_addr.s_addr = htonl (INADDR_ANY);
if (bind (sock, (struct sockaddr *) &name, sizeof (name)) < 0)
	{
	perror ("bind");
	return -1;
	}
     
return sock;
}/* end function make_socket */


int accept_connections(int argc, char **argv, int port, int l_coding_mode)
{
int a, c, i, j;
int sock;
fd_set active_fd_set, read_fd_set;
struct sockaddr_in clientname;
size_t size;
int new;
int bytes_to_be_send;
int bytes_send;
int bytes_written;
int bytes_read;
int bytes_in_frame;
int bytes_transmitted;
char inbuffer[HEADER_SIZE + 1];
int expected_length;
int error_flag;
unsigned char *ucbufferptr;
int total_frames_send;
time_t start_time, elapsed_time;
float frames_per_second;
unsigned long total_bytes_send;
float bytes_per_second;
int clients;
struct timeval timeout;
int message_number;
int said_no_clients_flag;
int confirm_msg_send_flag;
char temp[HEADER_SIZE];
char *ptr;

if(debug_flag)
	{
	fprintf(stderr, "accept_connections(): arg argc arv port=%d\n", port);
	}

/* Create the socket and set it up to accept connections. */
sock = make_socket (port);
if (listen (sock, MAX_LISTEN_QUEUE) < 0)
	{
	perror ("listen");
	return 0;
	}
     
/* Initialize the set of active sockets. */
FD_ZERO (&active_fd_set);
FD_SET (sock, &active_fd_set);

/* listen to user for commands */
FD_SET(fileno(stdin), &active_fd_set);     

/* report */
fprintf(stderr,\
"listening for connections on port %d\n\
lines starting with a '/' are commands\n\
/quit aborts server\n\
/stat shows statistics\n\
other text is send.\n\
After typing a line of text (max %d characters), wait for 'msg send'!\n",\
port, SERVER_STATUS - START_SERVER_TEXT);

/* test for connections and user action */
message_number = 0;
confirm_msg_send_flag = 0;
said_no_clients_flag = 0;
start_time = time(0);
elapsed_time = 0;
total_frames_send = 0;
total_bytes_send = 0;
clients = 0;
bytes_in_frame = 0;
frames_per_second = 0;
/* set for empty transmit text */
DataBuffer[START_SERVER_TEXT] = 0;
while (1)
	{
	/* Block until input arrives on one or more active sockets. */
	read_fd_set = active_fd_set;

	timeout.tv_sec = 1;
	timeout.tv_usec = 0;
	a = select (FD_SETSIZE, &read_fd_set, NULL, NULL, &timeout);
	if(a < 0)
		{
		perror ("select");
		return 0;
		}
	if(a == 0)/* timeout */     
		{
		/*fprintf(stderr, "server timeout\n");	*/
		/* basically we could return here and / or do something else */
		if(clients == 0)
			{
			if(! said_no_clients_flag)
				{
				fprintf(stderr, "No clients\n");
				said_no_clients_flag = 1;
				}
			}
		else
			{
			said_no_clients_flag = 0;
			}
		continue;
		}

	/* Service all the sockets with input pending. */
	for (i = 0; i < FD_SETSIZE; ++i)
		{
		if (FD_ISSET (i, &read_fd_set) )
			{
			/* test for user abort */
			if(i == fileno(stdin) )
				{
				/* clear tx header */
				for(j = 0; j < HEADER_SIZE ; j++)
					{
					DataBuffer[j] = 0;
					}
				a = read (i, temp, SERVER_STATUS - START_SERVER_TEXT);

				/* string termination */
				temp[a] = 0;
				if(debug_flag)
					{
					fprintf(stderr,\
					"read returned a=%d, temp=%s strlen(temp)=%d\n",\
					a, temp, strlen(temp) );
					}

				/* test for user commands */
				if(temp[0] != '/')
					{
					/* must be message */
					
					/* message to DataBuffer */
					strcpy(DataBuffer + START_SERVER_TEXT, temp);
				
					/* set the message number in the data to be transmitted */
					message_number++;
					DataBuffer[TEXT_MSG_NUMBER_HIGH] = message_number / 256;
					DataBuffer[TEXT_MSG_NUMBER_LOW] = message_number % 256;
					if(debug_flag)
						{
						fprintf(stderr,\
						"Sending: message_number=%d text=%s\n",\
						message_number, DataBuffer + START_SERVER_TEXT);
						}

 					confirm_msg_send_flag = 1;
 
 					continue;
					}/* end if text message */
				
				/* must be a command */				
				/* sort commands */
				if(strncasecmp(temp, "/quit", 5) == 0)
					{
					fprintf(stderr,\
					"\naccept_connections(): user abort.\n");

					return 0; 
					}/* end if /quit */
				else if(strncasecmp(temp, "/stat", 5) == 0)
					{
					fprintf(stderr,\
					"total frames=%d  f/s=%.2f  KB/s=%.2f  B/f=%d  clients=%d  \
messages send=%d\n",\
					total_frames_send,\
					frames_per_second, bytes_per_second / 1024.0,\
					bytes_in_frame, clients, message_number);
					
					continue;
					}/* end if /stat */
				else /* unknown command */
					{
					fprintf(stderr, "Unknown command\n");					

					continue;
					}
				/* end user commands */	
				
 				}/* end if something on stdin */

			if (i == sock) /* Connection request on original socket. */
				{
				size = sizeof (clientname);
				new = accept (sock, (struct sockaddr *) &clientname, &size);
				if (new < 0)
					{
					perror ("accept");
					return 0;
					}
				/* report connect */
				fprintf (stderr,\
				"Server: connect from host %s.\n",\
				inet_ntoa (clientname.sin_addr) );
/*this gives incorrect port why?	ntohs (clientname.sin_port));*/

				FD_SET (new, &active_fd_set);

				/* one client more */
				clients++;
				}
			else /* Data arriving on an already-connected socket. */
				{
				errno = 0;
				bytes_read = read (i, inbuffer, HEADER_SIZE);
				if(debug_flag)
					{
					fprintf(stderr,\
					"server read(): bytes read=%d\n",\
					bytes_read);
					}
				if (bytes_read < 0) /* read error */
					{
					perror ("read");
					if(errno == EAGAIN) continue;
					if(errno == EINTR) continue;

					/* drop this connection */
					FD_CLR (i, &active_fd_set);
					close(i);

					/* one client less */
					if(clients > 0) clients--;

					/* try next socket */
					continue;
					return 0;
					}
				else if (bytes_read == 0) /* EOF */
					{
					FD_CLR (i, &active_fd_set);

					close (i);

					/* one client less */
					if(clients > 0) clients--;

					/* try next socket */
					continue;
					}

				/* some statistics */
				elapsed_time = time(0) - start_time;

				/* prevent divide by zero if here within one second */
				if(elapsed_time == 0) elapsed_time = 1;
				
				frames_per_second =\
				(float)total_frames_send / (float)elapsed_time;

				bytes_per_second =\
				(float)total_bytes_send / (float)elapsed_time;

				/* test if this client has permission to control camera */
				if(debug_flag)
					{
					fprintf(stderr,\
					"server_control_password=%s\n", server_control_password);
					fprintf(stderr,\
					"server_password=%s\n", server_password);
					fprintf(stderr, "inbuffer=%s\n", inbuffer);
					}
					
				/* signal to this client, default no command permission */	
				DataBuffer[FLAGS] &= 255 - 1;/* reset bit 0 */
				
				/* check for correct or not needed control password */
				if( (strcmp(server_control_password, "no") == 0) ||\
				(strcmp(server_control_password, inbuffer) == 0) )/* right psw */
					{
					/* test for a camera command from the client */
					if(inbuffer[START_COMMAND] != 0)
						{
						execute_camera_command(\
						inbuffer + START_COMMAND, &coding_mode);
						}		
					
					/*
					signal back to this client he has command permission.
					*/
					DataBuffer[FLAGS] |= 1;/* set bit 0 */
					}/* end if server_control_password correct */
				else /* do normal password check */
					{
					if( (strcmp(server_password, "no") != 0) &&\
					(strcmp(server_password, inbuffer) != 0) )/* wrong psw */
						{
						fprintf(stderr, "Incorrect password %s from %s\n",\
						inbuffer, inet_ntoa (clientname.sin_addr) );

						/* drop this connection */
						close(i);
						FD_CLR (i, &active_fd_set);

						/* one client less */
						if(clients > 0) clients--;

						/* try next socket */
						continue;
						}/* end if password incorrect */
					}/* end if normal password check */

				/* test for a message from the client */
				/* text string starts here  */
				if(inbuffer[START_CLIENT_TEXT] != 0)
					{
					if(text_to_screen_flag)
						{
						/* text to screen text array */
						strncpy(screen_text,\
						inbuffer + START_CLIENT_TEXT, 50);

						/* cut of the LF at end */
						ptr = strchr(screen_text, '\n');
						if(ptr) *ptr = 0;
						else screen_text[45] = 0;/* if no LF, limit to 45 */
						}
					else /* text to console */	
						{
						/* note there is always a LF at the end */
						fprintf(stderr,\
						"Msg: %s", inbuffer + START_CLIENT_TEXT);
						}/* end if text to console */
					}/* end if text from client */

				/* get the camera status from cam header into tx header */
				camera_status_to_buffer(DataBuffer);
				
				/* server status to txbuffer */
				server_status_to_buffer(\
				DataBuffer, clients, bytes_per_second, frames_per_second);

				/*
				get frame from camera into gStreamBuffer
				(DataBuffer + HEADER_SIZE)
				*/
				bytes_in_frame = 0;
				if(! get_a_frame(&bytes_in_frame, coding_mode) )
					{
					fprintf(stderr, "Cannot get a frame\n");

					/* message to DataBuffer */
					strcpy(DataBuffer + START_SERVER_TEXT,\
					"COULD NOT GET A FRAME try /camera restart");
					/* of cause we could do it here automagically */			
					
					/* set the message number in the data to be transmitted */
					message_number++;
					DataBuffer[TEXT_MSG_NUMBER_HIGH] = message_number / 256;
					DataBuffer[TEXT_MSG_NUMBER_LOW] = message_number % 256;
					if(debug_flag)
						{
						fprintf(stderr,\
						"Sending: message_number=%d text=%s\n",\
						message_number, DataBuffer + START_SERVER_TEXT);
						}

 					confirm_msg_send_flag = 1;
					
					/* send only the header */
					bytes_in_frame = 0;
					}/* end get_a_frame failed */
				if(debug_flag)
					{
					fprintf(stderr,\
					"accept_internet_connection(): bytes_in_frame=%d\n",\
					bytes_in_frame);
					}

				/* write the handshake (number of bytes transmitted) */
				bytes_transmitted = HEADER_SIZE + bytes_in_frame;
				/* high byte first */
				DataBuffer[DATA_SIZE_VERY_HIGH] = bytes_transmitted / 65536;
				DataBuffer[DATA_SIZE_HIGH] = bytes_transmitted / 256;
				DataBuffer[DATA_SIZE_LOW] = bytes_transmitted % 256;

				/* try to write the DataBuffer */
				/*
				Note:
				The data buffer contains the HEADER_SIZE byte header,
				then the camera buffer (gStreamBuffer) at
				DataBuffer + HEADER_SIZE.
				The camera data starts with its own 64 byte buffer.
				*/
				/* add space for the handshake buffer */
				/*
				sending:
				mcam header + camera header + camera data.
				*/
				bytes_to_be_send = bytes_in_frame + HEADER_SIZE;
				bytes_send = 0;
				error_flag = 0;
				ucbufferptr = DataBuffer;
				while(1)
					{
					/* test if ready */
					if(debug_flag)
						{
						fprintf(stderr,\
				"accept_connections(): bytes_to_be_send=%d bytes_send=%d\n",\
						bytes_to_be_send, bytes_send);
						}

					if(bytes_send == bytes_to_be_send) break;

					/* write the data */
					errno = 0;
					bytes_written =\
					write(i, ucbufferptr, bytes_to_be_send - bytes_send);
					if(bytes_written < 0) /* error in write() */
						{
						perror("accept_connection(): write");
						
						/* try again */
						if(errno == EINTR) continue;
						if(errno == EAGAIN) continue;
						/* SIGPIPE is ignored in server mode */
							
						/* drop connection */
						FD_CLR (i, &active_fd_set);

						/* close socket */
						close(i);
							
						/* one client less */
						if(clients > 0) clients--;

						error_flag = 1;
						break;
						}/* end if write failed */
					bytes_send += bytes_written;

					ucbufferptr += bytes_written;

					if(debug_flag)
						{
						fprintf(stderr,
						"accept_connection(): %d of %d bytes written\n",\
						bytes_send, bytes_to_be_send);
						}
					}/* end while send data buffer */

				/* try next socket if error */
				if(error_flag) continue;

				total_frames_send++;
				total_bytes_send += (unsigned long)bytes_send;
				if(debug_flag)
					{
					fprintf(stderr,\
					"total_bytes_send=%lu\n", total_bytes_send);
					}
				if(confirm_msg_send_flag)
					{
					fprintf(stderr, "Msg send\n");
					confirm_msg_send_flag = 0;
					}
				
				}/* end if existing connection */
			}/* end if FD_ISSET */
		}/* end for server all connections */
	}/* end while wait for select */

return 1;
}/* end function accept_connections */


int execute_camera_command(BYTE __huge *buffer, int *l_coding_mode)
{
int brightness, contrast, saturation;
float fa, fb;

if(debug_flag)
	{
	fprintf(stderr, "execute_camera_command(): buffer=%lu\n",\
	buffer);
	}

if(! buffer) return 0;

/* test if any command */
if(*buffer == 0) return 0;

if(*buffer == HIGH_POWER)
	{
	fprintf(stderr, "Got command from client: highpower\n");
	CPIA_GotoHiPower();
	}
else if(*buffer == LOW_POWER)
	{
	fprintf(stderr, "Got command from client: lowpower\n");
	CPIA_GotoLoPower();
	}
else if(*buffer == RESET)
	{
	fprintf(stderr, "Got command from client: reset\n");
	CLINK_ForceWatchdogReset();
	}
else if(*buffer == COMPRESSION_ON)
	{
	fprintf(stderr, "Got command from client: compression on\n");
	CPIA_SetCompression(1, 0);
	}
else if(*buffer == COMPRESSION_OFF)
	{
	fprintf(stderr, "Got command from client: compression off\n");
	CPIA_SetCompression(0, 0);
	}
else if(*buffer == BIG_PICURE)
	{
	fprintf(stderr, "Got command from client: big (picture)\n");
/*	close_camera_window();*/

	/* global signal to other routines */
	picture_format = 1;

	CPIA_SetFormat(1, 1, 0);
	}
else if(*buffer == SMALL_PICURE)
	{
	fprintf(stderr, "Got command from client: small (picture)\n");
/*	close_camera_window();*/

	/* global signal to other routines */
	picture_format = 0;

	CPIA_SetFormat(0, 1, 0);
	}
else if(*buffer == CONTRAST)
	{
	fprintf(stderr, "Got command from client: contrast %d\n", buffer[1]);

	/* get brightness, contrast, saturation */
	CPIA_GetColourParams(\
	(BYTE *)&brightness, (BYTE *)&contrast, (BYTE *)&saturation );

	contrast = buffer[1] & 0xff;
	CPIA_SetColourParams(\
	(BYTE)brightness, (BYTE)contrast, (BYTE)saturation );
	}
else if(*buffer == BRIGHTNESS)
	{
	fprintf(stderr, "Got command from client: brightness %d\n", buffer[1]);

	/* get brightness, contrast, saturation */
	CPIA_GetColourParams(\
	(BYTE *)&brightness, (BYTE *)&contrast, (BYTE *)&saturation );

	brightness = buffer[1] & 0xff;
	CPIA_SetColourParams(\
	(BYTE)brightness, (BYTE)contrast, (BYTE)saturation );
	}
else if(*buffer == SATURATION)
	{
	fprintf(stderr, "Got command from client: saturation %d\n", buffer[1]);

	/* get brightness, contrast, saturation */
	CPIA_GetColourParams(\
	(BYTE *)&brightness, (BYTE *)&contrast, (BYTE *)&saturation );

	saturation = buffer[1] & 0xff;
	CPIA_SetColourParams(\
	(BYTE)brightness, (BYTE)contrast, (BYTE)saturation );
	}
else if(*buffer == AUTO_WHITE_BALANCE)
	{
	fprintf(stderr, "Got command from client: auto white balance\n");

	color_balance_mode = 1;
	CPIA_SetColourBalance(2, 0, 0, 0);	
	}
else if(*buffer == MAN_WHITE_BALANCE)
	{
	fprintf(stderr,\
"Got command from client: man white bal red=%d green=%d blue=%d\n",\
	buffer[1], buffer[2], buffer[3]);

	/* set manual */
	color_balance_mode = 0;
	CPIA_SetColourBalance(3, 0, 0, 0);
	
	/* set values */
	CPIA_SetColourBalance(\
	1,\
	(BYTE)buffer[1] & 0xff,\
	(BYTE)buffer[2] & 0xff,\
	(BYTE)buffer[3] & 0xff);
	}
else if(*buffer == AUTO_EXPOSURE)
	{
	fprintf(stderr, "Got command from client: auto exposure on\n");

	man_exposure_flag = 0;
	CPIA_SetExposure(0, 2, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0);
	}
else if(*buffer == MAN_EXPOSURE)
	{
	fprintf(stderr,\
	"Got command from client: man exposure\n\
	gain=%d\n\
 	fine exposure=%d\n\
 	coarse exposure=%d\n",\
	buffer[1], buffer[2], (buffer[3] * 256) + buffer[4]);

	/* set manual */
	man_exposure_flag = 1;
	CPIA_SetExposure(0, 3, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0);

	/* set values */
	/* note: coarse exposure sending high byte first! */
	CPIA_SetExposure(0, 1, 0, 0,\
	(BYTE)buffer[1] & 0xff,		/* gain */
	(BYTE)buffer[2] & 127,		/* exposure fine */
	(BYTE)buffer[4] & 0xff,		/* exposure coarse low */
	(BYTE)buffer[3] & 0xff,		/* exp coarse high */
	0, 0, 0, 0);
	}
else if(*buffer == GAIN_CEILING)
	{
	fprintf(stderr,\
	"Got command from client: gain_ceiling=%d\n", buffer[1]);

	gain_ceiling =  (BYTE)buffer[1];
	CPIA_SetExposure(\
	(BYTE)buffer[1] & 0xff, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0);
	}
else if(*buffer == COMPENSATION_GAIN)
	{
	fprintf(stderr, "Got command from client:\n\
	red_compensation=%d\n\
	green_1_compensation=%d\n\
	green_2_compensation=%d\n\
	blue_compensation=%d\n",\
	buffer[1], buffer[2], buffer[3], buffer[4]);

	CPIA_SetExposure(0, 0, 1, 0,  0, 0, 0, 0,\
	(BYTE)buffer[1] & 0xff,		/* red_compensation */
	(BYTE)buffer[2] & 0xff,		/* green_1_compensation */
	(BYTE)buffer[3] & 0xff,		/* green_2_compensation */	
	(BYTE)buffer[4] & 0xff);	/* blue_compensation */
	}
else if(*buffer == CAMERA_RESTART)
	{
	fprintf(stderr, "Got command from client: restart\n");
	
	restart_camera();
	}
else if(*buffer == NO_CODING)
	{
	fprintf(stderr, "Got command from client: coding off\n");

	coding_mode = NO_CODING;
	}
else if(*buffer == JPEG_CODING)
	{
	fprintf(stderr, "Got command from client: coding jpeg\n");

	coding_mode = JPEG_CODING;
	}
else if(*buffer == MPEG2_CODING)
	{
	fprintf(stderr, "Got command from client: coding mpeg\n");

	coding_mode = MPEG2_CODING;
	}
else if(*buffer == WAVELET_CODING)
	{
	fprintf(stderr, "Got command from client: coding wavelet\n");

	coding_mode = WAVELET_CODING;
	}
else if(*buffer == CODING_QUALITY)
	{
	fprintf(stderr, "Got command from client: coding_quality %d\n",\
	buffer[1]);
	coding_quality = buffer[1];
	}
else if(*buffer == CLOCK_ON)
	{
	fprintf(stderr, "Got command from client: clock on\n");
	show_time_flag = 1;
	}
else if(*buffer == CLOCK_OFF)
	{
	fprintf(stderr, "Got command from client: clock off\n");
	show_time_flag = 0;
	}
else if(*buffer == CLOCK_UTC)
	{
	fprintf(stderr, "Got command from client: use UTC time\n");
	use_utc_flag = 1;
	}
else if(*buffer == CLOCK_LOCAL)
	{
	fprintf(stderr, "Got command from client: use local time\n");
	use_utc_flag = 0;
	}
else if(*buffer == CLOCK_POSITION)
	{
	clock_position = *(buffer + 1);
	if(*(buffer + 1) == TOP_LEFT)
		{
		fprintf(stderr, "Got command from client: clock position top left\n");
		}
	if(*(buffer + 1) == TOP_MIDDLE)
		{
		fprintf(stderr, "Got command from client: clock position top middle\n");
		}
	if(*(buffer + 1) == TOP_RIGHT)
		{
		fprintf(stderr, "Got command from client: clock position top right\n");
		}
	if(*(buffer + 1) == BOTTOM_LEFT)
		{
		fprintf(stderr, "Got command from client: clock position bottom left\n");
		}
	if(*(buffer + 1) == BOTTOM_MIDDLE)
		{
		fprintf(stderr, "Got command from client: clock position bottom middle\n");
		}
	if(*(buffer + 1) == BOTTOM_RIGHT)
		{
		fprintf(stderr, "Got command from client: clock position bottom right\n");
		}
	}/* end if clock position */

else if(*buffer == TEXT_TO_SCREEN)
	{
	fprintf(stderr, "Got command from client: text to screen\n");
	if(debug_flag)
		{
		fprintf(stderr, "strlen(screen_text)=%d\n", strlen(screen_text) );
		}

	/* clear the array */
/*	screen_text[0] = 0;*/
	text_to_screen_flag = 1;
	}
else if(*buffer == TEXT_TO_CONSOLE)
	{
	fprintf(stderr, "Got command from client: text to console\n");

	/* leave the text in the array */

	text_to_screen_flag = 0;
	}

else if(*buffer == TEXT_POSITION)
	{
	text_x = (*(buffer + 1) ) * 256 + *(buffer + 2);

	text_y = (*(buffer + 3) ) * 256 + *(buffer + 4);

	fprintf(stderr, "Got command from client: text position\n\
	x=%d\n\
	y=%d\n",\
	text_x, text_y);	
	}

else if(*buffer == TEXT_COLOR)
	{
	text_fr = *(buffer + 1);
	text_fg = *(buffer + 2);
	text_fb = *(buffer + 3);
	text_br = *(buffer + 4);
	text_bg = *(buffer + 5);
	text_bb = *(buffer + 6);
	fprintf(stderr, "Got command from client: text color\n\
	fr=%d fg=%d fb=%d\n\
	br=%d bg=%d bb=%d\n",\
	text_fr, text_fg, text_fb, text_br, text_bg, text_bb);
	}

else if(*buffer == CHAR_SET_S14)
	{
	selected_font = S14_FONT;
	}
else if(*buffer == CHAR_SET_TXT)
	{
	selected_font = TXT_FONT;
	}
else if(*buffer == SET_APERTURE_CORRECTION)
	{
	apcor_values[0] = *(buffer + 1);
	apcor_values[1] = *(buffer + 2);
	apcor_values[2] = *(buffer + 3);
	apcor_values[3] = *(buffer + 4);
	CPIA_SetApcor(\
	apcor_values[0], apcor_values[1], apcor_values[2], apcor_values[3]);

	fprintf(stderr,\
	"Got command from client: apcor %d %d  %d %d  %d %d  %d %d\n",\
    (*(buffer + 1) & 0xf0) >> 4, *(buffer + 1) & 0xf,
	(*(buffer + 2) & 0xf0) >> 4, *(buffer + 2) & 0xf,
	(*(buffer + 3) & 0xf0) >> 4, *(buffer + 3) & 0xf,
	(*(buffer + 4) & 0xf0) >> 4, *(buffer + 4) & 0xf);
	}
else if(*buffer == SET_APCOR_DEFAULT)
	{
	apcor_values[0] = 0x2c;
	apcor_values[1] = 0x2a;
	apcor_values[2] = 0x4d;
	apcor_values[3] = 0x4a;
	CPIA_SetApcor(\
	apcor_values[0], apcor_values[1], apcor_values[2], apcor_values[3]);

	fprintf(stderr, "Got command from client: apcor_default\n");
	}
else if(*buffer == SET_SENSOR_FRAMES_PER_SECOND)
	{
	sensor_frames_per_second = *(buffer + 1);

	set_sensor_frames_per_second(sensor_frames_per_second, &fb);

	fprintf(stderr,\
	"Got command from client set sensor frames per second %d\n",\
	sensor_frames_per_second);
	}

/* NOTE: server status is always send */ 
else return 0;/* not a valid command */
}/* end functon execute_camera_command */


int restart_camera()
{
fprintf(stderr, "restart_camera(): trying watchdog reset\n");
CLINK_ForceWatchdogReset();

fprintf(stderr, "restart_camera(): trying lowpower\n");
CPIA_GotoLoPower();

fprintf(stderr, "restart_camera(): trying highpower\n");
CPIA_GotoHiPower();

return 1;
}/* end function restart_camera */


int camera_status_to_buffer(BYTE __huge *buffer)
{
BYTE __huge *bhptr;

bhptr = buffer + SERVER_STATUS;

/* get brightness, contrast, saturation */
CPIA_GetColourParams(\
(BYTE *)(bhptr + 11),		/* brightness */
(BYTE *)(bhptr + 12),		/* contrast */
(BYTE *)(bhptr + 13) );		/* saturation */

/* get current color gains */
CPIA_GetColourBalance(\
(BYTE *)(bhptr + 14),		/* current_red_gain */
(BYTE *)(bhptr + 15),		/* current_green_gain */
(BYTE *)(bhptr + 16) );		/* current_blue_gain */

/* get current exposure */
CPIA_GetExposure(\
(BYTE *)(bhptr + 17),		/* current_gain, is gain_ceiling - 1 */
(BYTE *)(bhptr + 18),		/* current_fine_exposure */
(BYTE *)(bhptr + 20),		/* NOTE reversal current_coarse_exposure_low */
(BYTE *)(bhptr + 19),		/* NOTE reversal current_coarse_exposure_high */
(BYTE *)(bhptr + 21),		/* red_compensation */
(BYTE *)(bhptr + 22),		/* green_1_compensation */
(BYTE *)(bhptr + 23),		/* green_2_compensation */
(BYTE *)(bhptr + 24) );		/* blue_compensation */

return 1;
}/* end function camera_status_to_buffer */


int server_status_to_buffer(\
BYTE __huge *buffer, int clients,\
float bytes_per_second, float frames_per_second)
{
int a;
float fa;
BYTE __huge *bhptr;
int major_server_version, minor_server_version;

if(debug_flag)
	{
	fprintf(stderr, "server_status_to_buffer():\n\
	arg clients=%d bytes_per_second=%.2f frames_per_second=%.2f\n",\
	clients, bytes_per_second, frames_per_second);
	}

bhptr = buffer + SERVER_STATUS;

/* number of clients connected to server */
*(bhptr + 35) = clients;

/* frames per 1000 seconds */
fa = frames_per_second * 1000.0;
a = (int) fa;
*(bhptr + 37) = a / 256;
*(bhptr + 38) = a % 256;

/* bytes per second, or if greater then 1024 kilobytes per second */
/* send KB if in KB */
if(bytes_per_second > 1024.0)
	{
	*(buffer + FLAGS) |= 4;
	fa = bytes_per_second / 1024.0;
	}
else
	{
	*(buffer + FLAGS) &= 255 - 4;
	fa = bytes_per_second;
	}
a = (int) fa;
*(bhptr + 39) = a / 256;
*(bhptr + 40) = a % 256;

/*
Using bit 1 for bit 0 coding_mode, and bit 3 for bit 1 coding flag
00 = no coding
01 = jpeg
10 = mpeg2
11 = wavelet
*/
if(coding_mode == NO_CODING)
	{
	/* 00 */
	*(buffer + FLAGS) &= 255 - 2;
	*(buffer + FLAGS) &= 255 - 8;
	}
else if(coding_mode == JPEG_CODING)
	{
	/* 01 */
	*(buffer + FLAGS) |= 2;
	*(buffer + FLAGS) &= 255 - 8;
	}
else if(coding_mode == MPEG2_CODING)
	{
	/* 10 */
	*(buffer + FLAGS) &= 255 - 2;
	*(buffer + FLAGS) |= 8;
	}
else if(coding_mode == WAVELET_CODING)
	{
	/* 11 */
	*(buffer + FLAGS) |= 2;
	*(buffer + FLAGS) |= 8;
	}

*(bhptr + 33) = coding_quality;

if(man_exposure_flag) *(buffer + FLAGS) |= 16;
else  *(buffer + FLAGS) &= 255 - 16;

if(color_balance_mode) *(buffer + FLAGS) |= 32;
else  *(buffer + FLAGS) &= 255 - 32;

if(show_time_flag) *(buffer + FLAGS) |= 64;
else *(buffer + FLAGS) &= 255 - 64;

if(use_utc_flag) *(buffer + FLAGS) |= 128;
else *(buffer + FLAGS) &= 255 - 128;

*(bhptr + 34) = clock_position;

if(picture_format == 1) *(buffer + FLAGS2) |= 1;
else *(buffer + FLAGS2) &= 255 - 1;

if(text_to_screen_flag)  *(buffer + FLAGS2) |= 2;
else *(buffer + FLAGS2) &= 255 - 2;

/* camera text x position */
*(bhptr + 25) = text_x % 256;
if(text_x / 256) *(buffer + FLAGS2) |= 4;
else *(buffer + FLAGS2) &= 255 - 4; 

/* camera text y position */
*(bhptr + 26) = text_y % 256;
if(text_y / 256) *(buffer + FLAGS2) |= 8;
else *(buffer + FLAGS2) &= 255 - 8; 

/* camera text color */
*(bhptr + 27) = text_fr;
*(bhptr + 28) = text_fg;
*(bhptr + 29) = text_fb;
*(bhptr + 30) = text_br;
*(bhptr + 31) = text_bg;
*(bhptr + 32) = text_bb;

if(selected_font == TXT_FONT) *(buffer + FLAGS2) |= 16; 
if(selected_font == S14_FONT) *(buffer + FLAGS2) &= 255 - 16;

/* apertur correction */
*(bhptr + 0) = apcor_values[0];
*(bhptr + 1) = apcor_values[1];
*(bhptr + 2) = apcor_values[2];
*(bhptr + 3) = apcor_values[3];

/* server frames per second */
*(bhptr + 10) = (int) sensor_frames_per_second;

sscanf(VERSION, "%d.%d", &major_server_version, &minor_server_version);

*(bhptr + 4) = (major_server_version << 4) + minor_server_version;

return 1;
}/* end function server_status_to_buffer */


